# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved.  See COPYING file
"""
Provide some useful helper function for pyverbs' tests.
"""
from itertools import combinations as com
from string import ascii_lowercase as al
import random

from pyverbs.qp import QPCap, QPInitAttrEx
import pyverbs.device as d
import pyverbs.enums as e

MAX_MR_SIZE = 4194304
# Some HWs limit DM address and length alignment to 4 for read and write
# operations. Use a minimal length and alignment that respect that.
# For creation purposes use random alignments. As this is log2 of address
# alignment, no need for large numbers.
MIN_DM_SIZE = 4
DM_ALIGNMENT = 4
MIN_DM_LOG_ALIGN = 0
MAX_DM_LOG_ALIGN = 6
# Raw Packet QP supports TSO header, which creates a larger send WQE.
MAX_RAW_PACKET_SEND_WR = 2500


def get_mr_length():
    """
    Provide a random value for MR length. We avoid large buffers as these
    allocations typically fails.
    We use random.random() instead of randrange() or randint() due to
    performance issues when generating very large pseudo random numbers.
    :return: A random MR length
    """
    return int(MAX_MR_SIZE * random.random())


def filter_illegal_access_flags(element):
    """
    Helper function to filter illegal access flags combinations
    :param element: A list of access flags to check
    :return: True if this list is legal, else False
    """
    if e.IBV_ACCESS_REMOTE_ATOMIC in element or e.IBV_ACCESS_REMOTE_WRITE:
        if e.IBV_ACCESS_LOCAL_WRITE:
            return False
    return True


def get_access_flags(ctx):
    """
    Provide an array of random legal access flags for an MR.
    Since remote write and remote atomic require local write permission, if
    one of them is randomly selected without local write, local write will be
    added as well.
    After verifying that the flags selection is legal, it is appended to an
    array, assuming it wasn't previously appended.
    :param ctx: Device Context to check capabilities
    :param num: Size of initial collection
    :return: A random legal value for MR flags
    """
    attr = ctx.query_device()
    attr_ex = ctx.query_device_ex()
    vals = list(e.ibv_access_flags)
    if not attr_ex.odp_caps.general_caps & e.IBV_ODP_SUPPORT:
        vals.remove(e.IBV_ACCESS_ON_DEMAND)
    if not attr.device_cap_flags & e.IBV_DEVICE_MEM_WINDOW:
        vals.remove(e.IBV_ACCESS_MW_BIND)
    if not attr.atomic_caps & e.IBV_ATOMIC_HCA:
        vals.remove(e.IBV_ACCESS_REMOTE_ATOMIC)
    arr = []
    for i in range(1, len(vals)):
        tmp = list(com(vals, i))
        tmp = filter(filter_illegal_access_flags, tmp)
        for t in tmp:  # Iterate legal combinations and bitwise OR them
            val = 0
            for flag in t:
                val += flag.value
            arr.append(val)
    return arr


def get_dm_attrs(dm_len):
    """
    Initializes an AllocDmAttr member with the given length and random
    alignment. It currently sets comp_mask = 0 since other comp_mask values
    are not supported.
    :param dm_len:
    :return: An initialized AllocDmAttr object
    """
    align = random.randint(MIN_DM_LOG_ALIGN, MAX_DM_LOG_ALIGN)
    return d.AllocDmAttr(dm_len, align, 0)


def sample(coll):
    """
    Returns a random-length subset of the given collection.
    :param coll: The collection to sample
    :return: A subset of <collection>
    """
    return random.sample(coll, int((len(coll) + 1) * random.random()))


def random_qp_cap(attr):
    """
    Initializes a QPCap object with valid values based on the device's
    attributes.
    It doesn't check the max WR limits since they're reported for smaller WR
    sizes.
    :return: A QPCap object
    """
    # We use significantly smaller values than those in device attributes.
    # The attributes reported by the device don't take into account possible
    # larger WQEs that include e.g. memory window.
    send_wr = random.randint(1, int(attr.max_qp_wr / 8))
    recv_wr = random.randint(1, int(attr.max_qp_wr / 8))
    send_sge = random.randint(1, int(attr.max_sge / 2))
    recv_sge = random.randint(1, int(attr.max_sge / 2))
    inline = random.randint(0, 16)
    return QPCap(send_wr, recv_wr, send_sge, recv_sge, inline)


def random_qp_create_mask(qpt, attr_ex):
    """
    Select a random sublist of ibv_qp_init_attr_mask. Some of the options are
    not yet supported by pyverbs and will not be returned. TSO support is
    checked for the device and the QP type. If it doesn't exist, TSO will not
    be set.
    :param qpt: Current QP type
    :param attr_ex: Extended device attributes for capability checks
    :return: A sublist of ibv_qp_init_attr_mask
    """
    has_tso = attr_ex.tso_caps.max_tso > 0 and \
        attr_ex.tso_caps.supported_qpts & 1 << qpt
    supp_flags = [e.IBV_QP_INIT_ATTR_CREATE_FLAGS,
                  e.IBV_QP_INIT_ATTR_MAX_TSO_HEADER]
    # Either PD or XRCD flag is needed, XRCD is not supported yet
    selected = sample(supp_flags)
    selected.append(e.IBV_QP_INIT_ATTR_PD)
    if e.IBV_QP_INIT_ATTR_MAX_TSO_HEADER in selected and not has_tso:
        selected.remove(e.IBV_QP_INIT_ATTR_MAX_TSO_HEADER)
    mask = 0
    for s in selected:
        mask += s.value
    return mask


def get_create_qp_flags_raw_packet(attr_ex):
    """
    Select random QP creation flags for Raw Packet QP. Filter out unsupported
    flags prior to selection.
    :param attr_ex: Device extended attributes to check capabilities
    :return: A random combination of QP creation flags
    """
    has_fcs = attr_ex.device_cap_flags_ex & e._IBV_DEVICE_RAW_SCATTER_FCS
    has_cvlan = attr_ex.raw_packet_caps & e.IBV_RAW_PACKET_CAP_CVLAN_STRIPPING
    has_padding = attr_ex.device_cap_flags_ex & \
        e._IBV_DEVICE_PCI_WRITE_END_PADDING
    l = list(e.ibv_qp_create_flags)
    l.remove(e.IBV_QP_CREATE_SOURCE_QPN)  # UD only
    if not has_fcs:
        l.remove(e.IBV_QP_CREATE_SCATTER_FCS)
    if not has_cvlan:
        l.remove(e.IBV_QP_CREATE_CVLAN_STRIPPING)
    if not has_padding:
        l.remove(e.IBV_QP_CREATE_PCI_WRITE_END_PADDING)
    flags = sample(l)
    val = 0
    for i in flags:
        val |= i.value
    return val


def random_qp_create_flags(qpt, attr_ex):
    """
    Select a random sublist of ibv_qp_create_flags according to the QP type.
    :param qpt: Current QP type
    :param attr_ex: Used for Raw Packet QP to check device capabilities
    :return: A sublist of ibv_qp_create_flags
    """
    if qpt == e.IBV_QPT_RAW_PACKET:
        return get_create_qp_flags_raw_packet(attr_ex)
    elif qpt == e.IBV_QPT_UD:
        # IBV_QP_CREATE_SOURCE_QPN is only supported by mlx5 driver and is not
        # to be check in unittests.
        return random.choice([0, 2])  # IBV_QP_CREATE_BLOCK_SELF_MCAST_LB
    else:
        return 0


def random_qp_init_attr_ex(attr_ex, attr, qpt=None):
    """
    Create a random-valued QPInitAttrEX object with the given QP type.
    QP type affects QP capabilities, so allow users to set it and still get
    valid attributes.
    :param attr_ex: Extended device attributes for capability checks
    :param attr: Device attributes for capability checks
    :param qpt: Requested QP type
    :return: A valid initialized QPInitAttrEx object
    """
    max_tso = 0
    if qpt is None:
        qpt = random.choice([e.IBV_QPT_RC, e.IBV_QPT_UC, e.IBV_QPT_UD,
                             e.IBV_QPT_RAW_PACKET])
    qp_cap = random_qp_cap(attr)
    if qpt == e.IBV_QPT_RAW_PACKET and \
       qp_cap.max_send_wr > MAX_RAW_PACKET_SEND_WR:
        qp_cap.max_send_wr = MAX_RAW_PACKET_SEND_WR
    sig = random.randint(0, 1)
    mask = random_qp_create_mask(qpt, attr_ex)
    if mask & e.IBV_QP_INIT_ATTR_CREATE_FLAGS:
        cflags = random_qp_create_flags(qpt, attr_ex)
    else:
        cflags = 0
    if mask & e.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
        if qpt != e.IBV_QPT_RAW_PACKET:
            mask -= e.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
        else:
            max_tso = \
                random.randint(16, int(attr_ex.tso_caps.max_tso / 400))
    qia = QPInitAttrEx(qp_type=qpt, cap=qp_cap, sq_sig_all=sig, comp_mask=mask,
                       create_flags=cflags, max_tso_header=max_tso)
    return qia



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019, Mellanox Technologies. All rights reserved.
import weakref

from pyverbs.pyverbs_error import PyverbsRDMAError, PyverbsError
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.device cimport Context, DM
from .mr cimport MR, MW, DMMR
from pyverbs.addr cimport AH
from pyverbs.qp cimport QP

cdef extern from 'errno.h':
    int errno


cdef class PD(PyverbsCM):
    def __cinit__(self, Context context not None):
        """
        Initializes a PD object. A reference for the creating Context is kept
        so that Python's GC will destroy the objects in the right order.
        :param context: The Context object creating the PD
        :return: The newly created PD on success
        """
        self.pd = v.ibv_alloc_pd(<v.ibv_context*>context.context)
        if self.pd == NULL:
            raise PyverbsRDMAErrno('Failed to allocate PD', errno)
        self.ctx = context
        context.add_ref(self)
        self.logger.debug('PD: Allocated ibv_pd')
        self.mrs = weakref.WeakSet()
        self.mws = weakref.WeakSet()
        self.ahs = weakref.WeakSet()
        self.qps = weakref.WeakSet()

    def __dealloc__(self):
        """
        Closes the inner PD.
        :return: None
        """
        self.close()

    cpdef close(self):
        """
        Closes the underlying C object of the PD.
        PD may be deleted directly or indirectly by closing its context, which
        leaves the Python PD object without the underlying C object, so during
        destruction, need to check whether or not the C object exists.
        :return: None
        """
        self.logger.debug('Closing PD')
        self.close_weakrefs([self.qps, self.ahs, self.mws, self.mrs])
        if self.pd != NULL:
            rc = v.ibv_dealloc_pd(self.pd)
            if rc != 0:
                raise PyverbsRDMAErrno('Failed to dealloc PD')
            self.pd = NULL
            self.ctx = None

    cdef add_ref(self, obj):
        if isinstance(obj, MR) or isinstance(obj, DMMR):
            self.mrs.add(obj)
        elif isinstance(obj, MW):
            self.mws.add(obj)
        elif isinstance(obj, AH):
            self.ahs.add(obj)
        elif isinstance(obj, QP):
            self.qps.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc . All rights reserved. See COPYING file

import unittest

import pyverbs.device as d

class PyverbsTestCase(unittest.TestCase):
    def setUp(self):
        """
        Opens the devices and queries them
        """
        lst = d.get_device_list()
        self.devices = []
        for dev in lst:
            c = d.Context(name=dev.name.decode())
            attr = c.query_device()
            attr_ex = c.query_device_ex()
            self.devices.append((c, attr, attr_ex))

    def tearDown(self):
        for tup in self.devices:
            tup[0].close()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2018, Mellanox Technologies. All rights reserved.

#cython: language_level=3


cdef extern from '<infiniband/verbs.h>':

    cpdef enum ibv_transport_type:
        IBV_TRANSPORT_UNKNOWN   = -1
        IBV_TRANSPORT_IB        = 0
        IBV_TRANSPORT_IWARP     = 1
        IBV_TRANSPORT_USNIC     = 2
        IBV_TRANSPORT_USNIC_UDP = 3

    cpdef enum ibv_node_type:
        IBV_NODE_UNKNOWN    = -1
        IBV_NODE_CA         = 1
        IBV_NODE_SWITCH     = 2
        IBV_NODE_ROUTER     = 3
        IBV_NODE_RNIC       = 4
        IBV_NODE_USNIC      = 5
        IBV_NODE_USNIC_UDP  = 6

    cpdef enum:
        IBV_LINK_LAYER_UNSPECIFIED
        IBV_LINK_LAYER_INFINIBAND
        IBV_LINK_LAYER_ETHERNET

    cpdef enum ibv_atomic_cap:
        IBV_ATOMIC_NONE
        IBV_ATOMIC_HCA
        IBV_ATOMIC_GLOB

    cpdef enum ibv_port_state:
        IBV_PORT_NOP                = 0
        IBV_PORT_DOWN               = 1
        IBV_PORT_INIT               = 2
        IBV_PORT_ARMED              = 3
        IBV_PORT_ACTIVE             = 4
        IBV_PORT_ACTIVE_DEFER       = 5

    cpdef enum ibv_port_cap_flags:
        IBV_PORT_SM                         = 1 <<  1
        IBV_PORT_NOTICE_SUP                 = 1 <<  2
        IBV_PORT_TRAP_SUP                   = 1 <<  3
        IBV_PORT_OPT_IPD_SUP                = 1 <<  4
        IBV_PORT_AUTO_MIGR_SUP              = 1 <<  5
        IBV_PORT_SL_MAP_SUP                 = 1 <<  6
        IBV_PORT_MKEY_NVRAM                 = 1 <<  7
        IBV_PORT_PKEY_NVRAM                 = 1 <<  8
        IBV_PORT_LED_INFO_SUP               = 1 <<  9
        IBV_PORT_SYS_IMAGE_GUID_SUP         = 1 << 11
        IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP  = 1 << 12
        IBV_PORT_EXTENDED_SPEEDS_SUP        = 1 << 14
        IBV_PORT_CAP_MASK2_SUP              = 1 << 15,
        IBV_PORT_CM_SUP                     = 1 << 16
        IBV_PORT_SNMP_TUNNEL_SUP            = 1 << 17
        IBV_PORT_REINIT_SUP                 = 1 << 18
        IBV_PORT_DEVICE_MGMT_SUP            = 1 << 19
        IBV_PORT_VENDOR_CLASS_SUP           = 1 << 20
        IBV_PORT_DR_NOTICE_SUP              = 1 << 21
        IBV_PORT_CAP_MASK_NOTICE_SUP        = 1 << 22
        IBV_PORT_BOOT_MGMT_SUP              = 1 << 23
        IBV_PORT_LINK_LATENCY_SUP           = 1 << 24
        IBV_PORT_CLIENT_REG_SUP             = 1 << 25
        IBV_PORT_IP_BASED_GIDS              = 1 << 26

    cpdef enum ibv_port_cap_flags2:
        IBV_PORT_SET_NODE_DESC_SUP              = 1 << 0
        IBV_PORT_INFO_EXT_SUP                   = 1 << 1
        IBV_PORT_VIRT_SUP                       = 1 << 2
        IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP    = 1 << 3
        IBV_PORT_LINK_WIDTH_2X_SUP              = 1 << 4
        IBV_PORT_LINK_SPEED_HDR_SUP             = 1 << 5

    cpdef enum ibv_mtu:
        IBV_MTU_256     = 1
        IBV_MTU_512     = 2
        IBV_MTU_1024    = 3
        IBV_MTU_2048    = 4
        IBV_MTU_4096    = 5

    cpdef enum ibv_event_type:
        IBV_EVENT_CQ_ERR
        IBV_EVENT_QP_FATAL
        IBV_EVENT_QP_REQ_ERR
        IBV_EVENT_QP_ACCESS_ERR
        IBV_EVENT_COMM_EST
        IBV_EVENT_SQ_DRAINED
        IBV_EVENT_PATH_MIG
        IBV_EVENT_PATH_MIG_ERR
        IBV_EVENT_DEVICE_FATAL
        IBV_EVENT_PORT_ACTIVE
        IBV_EVENT_PORT_ERR
        IBV_EVENT_LID_CHANGE
        IBV_EVENT_PKEY_CHANGE
        IBV_EVENT_SM_CHANGE
        IBV_EVENT_SRQ_ERR
        IBV_EVENT_SRQ_LIMIT_REACHED
        IBV_EVENT_QP_LAST_WQE_REACHED
        IBV_EVENT_CLIENT_REREGISTER
        IBV_EVENT_GID_CHANGE
        IBV_EVENT_WQ_FATAL

    cpdef enum ibv_access_flags:
        IBV_ACCESS_LOCAL_WRITE      = 1
        IBV_ACCESS_REMOTE_WRITE     = (1 << 1)
        IBV_ACCESS_REMOTE_READ      = (1 << 2)
        IBV_ACCESS_REMOTE_ATOMIC    = (1 << 3)
        IBV_ACCESS_MW_BIND          = (1 << 4)
        IBV_ACCESS_ZERO_BASED       = (1 << 5)
        IBV_ACCESS_ON_DEMAND        = (1 << 6)

    cpdef enum ibv_wr_opcode:
        IBV_WR_RDMA_WRITE
        IBV_WR_RDMA_WRITE_WITH_IMM
        IBV_WR_SEND
        IBV_WR_SEND_WITH_IMM
        IBV_WR_RDMA_READ
        IBV_WR_ATOMIC_CMP_AND_SWP
        IBV_WR_ATOMIC_FETCH_AND_ADD
        IBV_WR_LOCAL_INV
        IBV_WR_BIND_MW
        IBV_WR_SEND_WITH_INV
        IBV_WR_TSO

    cpdef enum ibv_send_flags:
        IBV_SEND_FENCE      = 1 << 0
        IBV_SEND_SIGNALED   = 1 << 1
        IBV_SEND_SOLICITED  = 1 << 2
        IBV_SEND_INLINE     = 1 << 3
        IBV_SEND_IP_CSUM    = 1 << 4

    cpdef enum ibv_qp_type:
        IBV_QPT_RC          = 2
        IBV_QPT_UC          = 3
        IBV_QPT_UD          = 4
        IBV_QPT_RAW_PACKET  = 8
        IBV_QPT_XRC_SEND    = 9
        IBV_QPT_XRC_RECV    = 10
        IBV_QPT_DRIVER      = 0xff

    cpdef enum ibv_qp_state:
        IBV_QPS_RESET
        IBV_QPS_INIT
        IBV_QPS_RTR
        IBV_QPS_RTS
        IBV_QPS_SQD
        IBV_QPS_SQE
        IBV_QPS_ERR
        IBV_QPS_UNKNOWN

    cpdef enum ibv_mw_type:
        IBV_MW_TYPE_1   = 1
        IBV_MW_TYPE_2   = 2

    cpdef enum ibv_wc_status:
        IBV_WC_SUCCESS
        IBV_WC_LOC_LEN_ERR
        IBV_WC_LOC_QP_OP_ERR
        IBV_WC_LOC_EEC_OP_ERR
        IBV_WC_LOC_PROT_ERR
        IBV_WC_WR_FLUSH_ERR
        IBV_WC_MW_BIND_ERR
        IBV_WC_BAD_RESP_ERR
        IBV_WC_LOC_ACCESS_ERR
        IBV_WC_REM_INV_REQ_ERR
        IBV_WC_REM_ACCESS_ERR
        IBV_WC_REM_OP_ERR
        IBV_WC_RETRY_EXC_ERR
        IBV_WC_RNR_RETRY_EXC_ERR
        IBV_WC_LOC_RDD_VIOL_ERR
        IBV_WC_REM_INV_RD_REQ_ERR
        IBV_WC_REM_ABORT_ERR
        IBV_WC_INV_EECN_ERR
        IBV_WC_INV_EEC_STATE_ERR
        IBV_WC_FATAL_ERR
        IBV_WC_RESP_TIMEOUT_ERR
        IBV_WC_GENERAL_ERR

    cpdef enum ibv_wc_opcode:
        IBV_WC_SEND
        IBV_WC_RDMA_WRITE
        IBV_WC_RDMA_READ
        IBV_WC_COMP_SWAP
        IBV_WC_FETCH_ADD
        IBV_WC_BIND_MW
        IBV_WC_LOCAL_INV
        IBV_WC_TSO
        IBV_WC_RECV         = 1 << 7
        IBV_WC_RECV_RDMA_WITH_IMM

    cpdef enum ibv_create_cq_wc_flags:
        IBV_WC_EX_WITH_BYTE_LEN                         = 1 << 0
        IBV_WC_EX_WITH_IMM                              = 1 << 1
        IBV_WC_EX_WITH_QP_NUM                           = 1 << 2
        IBV_WC_EX_WITH_SRC_QP                           = 1 << 3
        IBV_WC_EX_WITH_SLID                             = 1 << 4
        IBV_WC_EX_WITH_SL                               = 1 << 5
        IBV_WC_EX_WITH_DLID_PATH_BITS                   = 1 << 6
        IBV_WC_EX_WITH_COMPLETION_TIMESTAMP             = 1 << 7
        IBV_WC_EX_WITH_CVLAN                            = 1 << 8
        IBV_WC_EX_WITH_FLOW_TAG                         = 1 << 9
        IBV_WC_EX_WITH_TM_INFO                          = 1 << 10
        IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK   = 1 << 11

    cpdef enum ibv_wc_flags:
        IBV_WC_GRH              = 1 << 0
        IBV_WC_WITH_IMM         = 1 << 1
        IBV_WC_IP_CSUM_OK       = 1 << 2
        IBV_WC_WITH_INV         = 1 << 3
        IBV_WC_TM_SYNC_REQ      = 1 << 4
        IBV_WC_TM_MATCH         = 1 << 5
        IBV_WC_TM_DATA_VALID    = 1 << 6

    cpdef enum ibv_tm_cap_flags:
        IBV_TM_CAP_RC       = 1 << 0,

    cpdef enum ibv_srq_attr_mask:
        IBV_SRQ_MAX_WR      = 1 << 0,
        IBV_SRQ_LIMIT       = 1 << 1

    cpdef enum ibv_srq_type:
        IBV_SRQT_BASIC
        IBV_SRQT_XRC
        IBV_SRQT_TM

    cpdef enum ibv_srq_init_attr_mask:
        IBV_SRQ_INIT_ATTR_TYPE      = 1 << 0
        IBV_SRQ_INIT_ATTR_PD        = 1 << 1
        IBV_SRQ_INIT_ATTR_XRCD      = 1 << 2
        IBV_SRQ_INIT_ATTR_CQ        = 1 << 3
        IBV_SRQ_INIT_ATTR_TM        = 1 << 4

    cpdef enum ibv_mig_state:
        IBV_MIG_MIGRATED
        IBV_MIG_REARM
        IBV_MIG_ARMED

    cpdef enum ibv_qp_init_attr_mask:
        IBV_QP_INIT_ATTR_PD             = 1 << 0
        IBV_QP_INIT_ATTR_XRCD           = 1 << 1
        IBV_QP_INIT_ATTR_CREATE_FLAGS   = 1 << 2
        IBV_QP_INIT_ATTR_MAX_TSO_HEADER = 1 << 3
        IBV_QP_INIT_ATTR_IND_TABLE      = 1 << 4
        IBV_QP_INIT_ATTR_RX_HASH        = 1 << 5

    cpdef enum ibv_qp_create_flags:
        IBV_QP_CREATE_BLOCK_SELF_MCAST_LB   = 1 << 1
        IBV_QP_CREATE_SCATTER_FCS           = 1 << 8
        IBV_QP_CREATE_CVLAN_STRIPPING       = 1 << 9
        IBV_QP_CREATE_SOURCE_QPN            = 1 << 10
        IBV_QP_CREATE_PCI_WRITE_END_PADDING = 1 << 11

    cpdef enum ibv_qp_attr_mask:
        IBV_QP_STATE                = 1 << 0
        IBV_QP_CUR_STATE            = 1 << 1
        IBV_QP_EN_SQD_ASYNC_NOTIFY  = 1 << 2
        IBV_QP_ACCESS_FLAGS         = 1 << 3
        IBV_QP_PKEY_INDEX           = 1 << 4
        IBV_QP_PORT                 = 1 << 5
        IBV_QP_QKEY                 = 1 << 6
        IBV_QP_AV                   = 1 << 7
        IBV_QP_PATH_MTU             = 1 << 8
        IBV_QP_TIMEOUT              = 1 << 9
        IBV_QP_RETRY_CNT            = 1 << 10
        IBV_QP_RNR_RETRY            = 1 << 11
        IBV_QP_RQ_PSN               = 1 << 12
        IBV_QP_MAX_QP_RD_ATOMIC     = 1 << 13
        IBV_QP_ALT_PATH             = 1 << 14
        IBV_QP_MIN_RNR_TIMER        = 1 << 15
        IBV_QP_SQ_PSN               = 1 << 16
        IBV_QP_MAX_DEST_RD_ATOMIC   = 1 << 17
        IBV_QP_PATH_MIG_STATE       = 1 << 18
        IBV_QP_CAP                  = 1 << 19
        IBV_QP_DEST_QPN             = 1 << 20
        IBV_QP_RATE_LIMIT           = 1 << 25

    cpdef enum ibv_wq_type:
        IBV_WQT_RQ

    cpdef enum ibv_wq_init_attr_mask:
        IBV_WQ_INIT_ATTR_FLAGS      = 1 << 0

    cpdef enum ibv_wq_flags:
        IBV_WQ_FLAGS_CVLAN_STRIPPING        = 1 << 0
        IBV_WQ_FLAGS_SCATTER_FCS            = 1 << 1
        IBV_WQ_FLAGS_DELAY_DROP             = 1 << 2
        IBV_WQ_FLAGS_PCI_WRITE_END_PADDING  = 1 << 3

    cpdef enum ibv_wq_state:
        IBV_WQS_RESET
        IBV_WQS_RDY
        IBV_WQS_ERR
        IBV_WQS_UNKNOWN

    cpdef enum ibv_wq_attr_mask:
        IBV_WQ_ATTR_STATE       = 1 << 0
        IBV_WQ_ATTR_CURR_STATE  = 1 << 1
        IBV_WQ_ATTR_FLAGS       = 1 << 2

    cpdef enum ibv_rx_hash_function_flags:
        IBV_RX_HASH_FUNC_TOEPLITZ   = 1 << 0

    cpdef enum ibv_rx_hash_fields:
        IBV_RX_HASH_SRC_IPV4        = 1 << 0
        IBV_RX_HASH_DST_IPV4        = 1 << 1
        IBV_RX_HASH_SRC_IPV6        = 1 << 2
        IBV_RX_HASH_DST_IPV6        = 1 << 3
        IBV_RX_HASH_SRC_PORT_TCP    = 1 << 4
        IBV_RX_HASH_DST_PORT_TCP    = 1 << 5
        IBV_RX_HASH_SRC_PORT_UDP    = 1 << 6
        IBV_RX_HASH_DST_PORT_UDP    = 1 << 7

    cpdef enum ibv_ops_wr_opcode:
        IBV_WR_TAG_ADD
        IBV_WR_TAG_DEL
        IBV_WR_TAG_SYNC

    cpdef enum ibv_ops_flags:
        IBV_OPS_SIGNALED            = 1 << 0
        IBV_OPS_TM_SYNC             = 1 << 1

    cpdef enum ibv_flow_flags:
        IBV_FLOW_ATTR_FLAGS_ALLOW_LOOP_BACK = 1 << 0
        IBV_FLOW_ATTR_FLAGS_DONT_TRAP       = 1 << 1
        IBV_FLOW_ATTR_FLAGS_EGRESS          = 1 << 2

    cpdef enum ibv_flow_attr_type:
        IBV_FLOW_ATTR_NORMAL      = 0x0
        IBV_FLOW_ATTR_ALL_DEFAULT = 0x1
        IBV_FLOW_ATTR_MC_DEFAULT  = 0x2
        IBV_FLOW_ATTR_SNIFFER     = 0x3

    cpdef enum ibv_flow_spec_type:
        IBV_FLOW_SPEC_ETH           = 0x20
        IBV_FLOW_SPEC_IPV4          = 0x30
        IBV_FLOW_SPEC_IPV6          = 0x31
        IBV_FLOW_SPEC_IPV4_EXT      = 0x32
        IBV_FLOW_SPEC_ESP           = 0x34
        IBV_FLOW_SPEC_TCP           = 0x40
        IBV_FLOW_SPEC_UDP           = 0x41
        IBV_FLOW_SPEC_VXLAN_TUNNEL  = 0x50
        IBV_FLOW_SPEC_GRE           = 0x51
        IBV_FLOW_SPEC_MPLS          = 0x60
        IBV_FLOW_SPEC_INNER         = 0x100
        IBV_FLOW_SPEC_ACTION_TAG    = 0x1000
        IBV_FLOW_SPEC_ACTION_DROP   = 0x1001
        IBV_FLOW_SPEC_ACTION_HANDLE = 0x1002
        IBV_FLOW_SPEC_ACTION_COUNT  = 0x1003

    cpdef enum:
        IBV_QPF_GRH_REQUIRED

    cpdef enum ibv_counter_description:
        IBV_COUNTER_PACKETS
        IBV_COUNTER_BYTES

    cpdef enum ibv_read_counters_flags:
        IBV_READ_COUNTERS_ATTR_PREFER_CACHED = 1 << 0

    cpdef enum ibv_cq_init_attr_mask:
        IBV_CQ_INIT_ATTR_MASK_FLAGS = 1 << 0

    cpdef enum ibv_create_cq_attr_flags:
        IBV_CREATE_CQ_ATTR_SINGLE_THREADED = 1 << 0
        IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN  = 1 << 1

    cpdef enum ibv_odp_general_caps:
        IBV_ODP_SUPPORT             = 1
        IBV_ODP_SUPPORT_IMPLICIT    = 1 << 1

    cpdef enum ibv_odp_transport_cap_bits:
        IBV_ODP_SUPPORT_SEND        = 1 << 0
        IBV_ODP_SUPPORT_RECV        = 1 << 1
        IBV_ODP_SUPPORT_WRITE       = 1 << 2
        IBV_ODP_SUPPORT_READ        = 1 << 3
        IBV_ODP_SUPPORT_ATOMIC      = 1 << 4
        IBV_ODP_SUPPORT_SRQ_RECV    = 1 << 5

    cpdef enum ibv_device_cap_flags:
        IBV_DEVICE_RESIZE_MAX_WR            = 1 <<  0
        IBV_DEVICE_BAD_PKEY_CNTR            = 1 <<  1
        IBV_DEVICE_BAD_QKEY_CNTR            = 1 <<  2
        IBV_DEVICE_RAW_MULTI                = 1 <<  3
        IBV_DEVICE_AUTO_PATH_MIG            = 1 <<  4
        IBV_DEVICE_CHANGE_PHY_PORT          = 1 <<  5
        IBV_DEVICE_UD_AV_PORT_ENFORCE       = 1 <<  6
        IBV_DEVICE_CURR_QP_STATE_MOD        = 1 <<  7
        IBV_DEVICE_SHUTDOWN_PORT            = 1 <<  8
        IBV_DEVICE_INIT_TYPE                = 1 <<  9
        IBV_DEVICE_PORT_ACTIVE_EVENT        = 1 << 10
        IBV_DEVICE_SYS_IMAGE_GUID           = 1 << 11
        IBV_DEVICE_RC_RNR_NAK_GEN           = 1 << 12
        IBV_DEVICE_SRQ_RESIZE               = 1 << 13
        IBV_DEVICE_N_NOTIFY_CQ              = 1 << 14
        IBV_DEVICE_MEM_WINDOW               = 1 << 17
        IBV_DEVICE_UD_IP_CSUM               = 1 << 18
        IBV_DEVICE_XRC                      = 1 << 20
        IBV_DEVICE_MEM_MGT_EXTENSIONS       = 1 << 21
        IBV_DEVICE_MEM_WINDOW_TYPE_2A       = 1 << 23
        IBV_DEVICE_MEM_WINDOW_TYPE_2B       = 1 << 24
        IBV_DEVICE_RC_IP_CSUM               = 1 << 25
        IBV_DEVICE_RAW_IP_CSUM              = 1 << 26
        IBV_DEVICE_MANAGED_FLOW_STEERING    = 1 << 29

    cpdef enum ibv_raw_packet_caps:
        IBV_RAW_PACKET_CAP_CVLAN_STRIPPING  = 1 << 0
        IBV_RAW_PACKET_CAP_SCATTER_FCS      = 1 << 1
        IBV_RAW_PACKET_CAP_IP_CSUM          = 1 << 2
        IBV_RAW_PACKET_CAP_DELAY_DROP       = 1 << 3

    cdef unsigned long long IBV_DEVICE_RAW_SCATTER_FCS
    cdef unsigned long long IBV_DEVICE_PCI_WRITE_END_PADDING


cdef extern from "<infiniband/tm_types.h>":
    cpdef enum ibv_tmh_op:
        IBV_TMH_NO_TAG        = 0
        IBV_TMH_RNDV          = 1
        IBV_TMH_FIN           = 2
        IBV_TMH_EAGER         = 3

_IBV_DEVICE_RAW_SCATTER_FCS = IBV_DEVICE_RAW_SCATTER_FCS
_IBV_DEVICE_PCI_WRITE_END_PADDING = IBV_DEVICE_PCI_WRITE_END_PADDING



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file
"""
Test module for pyverbs' qp module.
"""
import random

from pyverbs.tests.base import PyverbsTestCase
from pyverbs.qp import QPInitAttr, QPAttr, QP
import pyverbs.tests.utils as u
import pyverbs.enums as e
from pyverbs.pd import PD
from pyverbs.cq import CQ


class QPTest(PyverbsTestCase):
    """
    Test various functionalities of the QP class.
    """

    def test_create_qp_no_attr_connected(self):
        """
        Test QP creation via ibv_create_qp without a QPAttr object proivded.
        QP type can be either RC or UC.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    qia = get_qp_init_attr(cq, [e.IBV_QPT_RC, e.IBV_QPT_UC],
                                           attr)
                    with QP(pd, qia) as qp:
                        assert qp.qp_state == e.IBV_QPS_RESET

    def test_create_qp_no_attr(self):
        """
        Test QP creation via ibv_create_qp without a QPAttr object proivded.
        QP type can be either Raw Packet or UD.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    for i in range(1, attr.phys_port_cnt + 1):
                        qpts = [e.IBV_QPT_UD, e.IBV_QPT_RAW_PACKET] \
                            if is_eth(ctx, i) else [e.IBV_QPT_UD]
                        qia = get_qp_init_attr(cq, qpts, attr)
                        with QP(pd, qia) as qp:
                            assert qp.qp_state == e.IBV_QPS_RESET

    def test_create_qp_with_attr_connected(self):
        """
        Test QP creation via ibv_create_qp without a QPAttr object proivded.
        QP type can be either RC or UC.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    qia = get_qp_init_attr(cq, [e.IBV_QPT_RC, e.IBV_QPT_UC],
                                           attr)
                    with QP(pd, qia, QPAttr()) as qp:
                        assert qp.qp_state == e.IBV_QPS_INIT

    def test_create_qp_with_attr(self):
        """
        Test QP creation via ibv_create_qp with a QPAttr object proivded.
        QP type can be either Raw Packet or UD.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    for i in range(1, attr.phys_port_cnt + 1):
                        qpts = [e.IBV_QPT_UD, e.IBV_QPT_RAW_PACKET] \
                            if is_eth(ctx, i) else [e.IBV_QPT_UD]
                        qia = get_qp_init_attr(cq, qpts, attr)
                        with QP(pd, qia, QPAttr()) as qp:
                            assert qp.qp_state == e.IBV_QPS_RTS

    def test_create_qp_ex_no_attr_connected(self):
        """
        Test QP creation via ibv_create_qp_ex without a QPAttr object proivded.
        QP type can be either RC or UC.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    qia = get_qp_init_attr_ex(cq, pd, [e.IBV_QPT_RC,
                                                       e.IBV_QPT_UC],
                                              attr, attr_ex)
                    with QP(ctx, qia) as qp:
                        assert qp.qp_state == e.IBV_QPS_RESET

    def test_create_qp_ex_no_attr(self):
        """
        Test QP creation via ibv_create_qp_ex without a QPAttr object proivded.
        QP type can be either Raw Packet or UD.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    for i in range(1, attr.phys_port_cnt + 1):
                        qpts = [e.IBV_QPT_UD, e.IBV_QPT_RAW_PACKET] \
                            if is_eth(ctx, i) else [e.IBV_QPT_UD]
                        qia = get_qp_init_attr_ex(cq, pd, qpts, attr,
                                                  attr_ex)
                        with QP(ctx, qia) as qp:
                            assert qp.qp_state == e.IBV_QPS_RESET

    def test_create_qp_ex_with_attr_connected(self):
        """
        Test QP creation via ibv_create_qp_ex with a QPAttr object proivded.
        QP type can be either RC or UC.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    qia = get_qp_init_attr_ex(cq, pd, [e.IBV_QPT_RC,
                                                       e.IBV_QPT_UC],
                                              attr, attr_ex)
                    with QP(ctx, qia, QPAttr()) as qp:
                        assert qp.qp_state == e.IBV_QPS_INIT

    def test_create_qp_ex_with_attr(self):
        """
        Test QP creation via ibv_create_qp_ex with a QPAttr object proivded.
        QP type can be either Raw Packet or UD.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    for i in range(1, attr.phys_port_cnt + 1):
                        qpts = [e.IBV_QPT_UD, e.IBV_QPT_RAW_PACKET] \
                            if is_eth(ctx, i) else [e.IBV_QPT_UD]
                        qia = get_qp_init_attr_ex(cq, pd, qpts, attr,
                                                  attr_ex)
                        with QP(ctx, qia, QPAttr()) as qp:
                            assert qp.qp_state == e.IBV_QPS_RTS

    def test_query_qp(self):
        """
        Queries a QP after creation. Verifies that its properties are as
        expected.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    for i in range(1, attr.phys_port_cnt + 1):
                        qpts = get_qp_types(ctx, i)
                        is_ex = random.choice([True, False])
                        if is_ex:
                            qia = get_qp_init_attr_ex(cq, pd, qpts, attr,
                                                      attr_ex)
                        else:
                            qia = get_qp_init_attr(cq, qpts, attr)
                        caps = qia.cap  # Save them to verify values later
                        qp = QP(ctx, qia) if is_ex else QP(pd, qia)
                        attr, init_attr = qp.query(e.IBV_QP_CUR_STATE |
                                                   e.IBV_QP_CAP)
                        verify_qp_attrs(caps, e.IBV_QPS_RESET, init_attr,
                                        attr)

    def test_modify_qp(self):
        """
        Queries a QP after calling modify(). Verifies that its properties are
        as expected.
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with CQ(ctx, 100, None, None, 0) as cq:
                    is_ex = random.choice([True, False])
                    if is_ex:
                        qia = get_qp_init_attr_ex(cq, pd, [e.IBV_QPT_UD],
                                                  attr, attr_ex)
                    else:
                        qia = get_qp_init_attr(cq, [e.IBV_QPT_UD], attr)
                    qp = QP(ctx, qia) if is_ex \
                        else QP(pd, qia)
                    qa = QPAttr()
                    qa.qkey = 0x123
                    qp.to_init(qa)
                    attr, iattr = qp.query(e.IBV_QP_QKEY)
                    assert attr.qkey == qa.qkey
                    qp.to_rtr(qa)
                    qa.sq_psn = 0x45
                    qp.to_rts(qa)
                    attr, iattr = qp.query(e.IBV_QP_SQ_PSN)
                    assert attr.sq_psn == qa.sq_psn
                    qa.qp_state = e.IBV_QPS_RESET
                    qp.modify(qa, e.IBV_QP_STATE)
                    assert qp.qp_state == e.IBV_QPS_RESET


def get_qp_types(ctx, port_num):
    """
    Returns a list of the commonly used QP types. Raw Packet QP will not be
    included if link layer is not Ethernet.
    :param ctx: The device's Context, to query the port's link layer
    :param port_num: Port number to query
    :return: An array of QP types that can be created on this port
    """
    qpts = [e.IBV_QPT_RC, e.IBV_QPT_UC, e.IBV_QPT_UD]
    if is_eth(ctx, port_num):
        qpts.append(e.IBV_QPT_RAW_PACKET)
    return qpts


def verify_qp_attrs(orig_cap, state, init_attr, attr):
    assert state == attr.cur_qp_state
    assert orig_cap.max_send_wr <= init_attr.cap.max_send_wr
    assert orig_cap.max_recv_wr <= init_attr.cap.max_recv_wr
    assert orig_cap.max_send_sge <= init_attr.cap.max_send_sge
    assert orig_cap.max_recv_sge <= init_attr.cap.max_recv_sge
    assert orig_cap.max_inline_data <= init_attr.cap.max_inline_data


def get_qp_init_attr(cq, qpts, attr):
    """
    Creates a QPInitAttr object with a QP type of the provided <qpts> array and
    other random values.
    :param cq: CQ to be used as send and receive CQ
    :param qpts: An array of possible QP types to use
    :param attr: Device attributes for capability checks
    :return: An initialized QPInitAttr object
    """
    qp_cap = u.random_qp_cap(attr)
    qpt = random.choice(qpts)
    sig = random.randint(0, 1)
    return QPInitAttr(qp_type=qpt, scq=cq, rcq=cq, cap=qp_cap, sq_sig_all=sig)


def get_qp_init_attr_ex(cq, pd, qpts, attr, attr_ex):
    """
    Creates a QPInitAttrEx object with a QP type of the provided <qpts> array
    and other random values.
    :param cq: CQ to be used as send and receive CQ
    :param pd: A PD object to use
    :param qpts: An array of possible QP types to use
    :param attr: Device attributes for capability checks
    :param attr_ex: Extended device attributes for capability checks
    :return: An initialized QPInitAttrEx object
    """
    qpt = random.choice(qpts)
    qia = u.random_qp_init_attr_ex(attr_ex, attr, qpt)
    qia.send_cq = cq
    qia.recv_cq = cq
    qia.pd = pd  # Only XRCD can be created without a PD
    return qia


def is_eth(ctx, port_num):
    """
    Querires the device's context's <port_num> port for its link layer.
    :param ctx: The Context to query
    :param port_num: Which Context's port to query
    :return: True if the port's link layer is Ethernet, else False
    """
    return ctx.query_port(port_num).link_layer == e.IBV_LINK_LAYER_ETHERNET



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019, Mellanox Technologies. All rights reserved.

import logging
from pyverbs.pyverbs_error import PyverbsRDMAError


cdef extern from 'errno.h':
    int errno

cpdef PyverbsRDMAErrno(str msg):
    return PyverbsRDMAError(msg, errno)

LOG_LEVEL=logging.INFO
LOG_FORMAT='[%(levelname)s] %(asctime)s %(filename)s:%(lineno)s: %(message)s'
logging.basicConfig(format=LOG_FORMAT, level=LOG_LEVEL, datefmt='%d %b %Y %H:%M:%S')

cdef class PyverbsObject(object):

    def __cinit__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def set_log_level(self, val):
        self.logger.setLevel(val)

    def close_weakrefs(self, iterables):
        """
        For each iterable element of iterables, pop each element and
        call its close() method. This method is used when an object is being
        closed while other objects still hold C references to it; the object
        holds weakrefs to such other object, and closes them before trying to
        teardown the C resources.
        :param iterables: an array of WeakSets
        :return: None
        """
        # None elements can be present if an object's close() was called more
        # than once (e.g. GC and by another object)
        for it in iterables:
            if it is None:
                continue
            while True:
                try:
                    tmp = it.pop()
                    tmp.close()
                except KeyError: # popping an empty set
                    break


cdef class PyverbsCM(PyverbsObject):
    """
    This is a base class for pyverbs' context manager objects. It includes
    __enter__ and __exit__ functions.
    close() is also declared but it should be overridden by each inheriting
    class.
    """
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.close()

    cpdef close(self):
        pass



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2018, Mellanox Technologies. All rights reserved. See COPYING file

"""
Device module introduces the Context and DeviceAttr class.
It allows user to open an IB device (using Context(name=<name>) and query it,
which returns a DeviceAttr object.
"""
import weakref

from .pyverbs_error import PyverbsRDMAError, PyverbsError
from pyverbs.cq cimport CQEX, CQ, CompChannel
from .pyverbs_error import PyverbsUserError
from pyverbs.base import PyverbsRDMAErrno
cimport pyverbs.libibverbs_enums as e
cimport pyverbs.libibverbs as v
from pyverbs.addr cimport GID
from pyverbs.mr import DMMR
from pyverbs.pd cimport PD
from pyverbs.qp cimport QP

cdef extern from 'errno.h':
    int errno
cdef extern from 'endian.h':
    unsigned long be64toh(unsigned long host_64bits);
cdef extern from 'stdlib.h':
    void free(void *ptr)
cdef extern from 'string.h':
    void *memset(void *s, int c, size_t n)
cdef extern from 'stdlib.h':
    void *malloc(size_t size)
cdef extern from 'stdint.h':
    ctypedef int uint64_t


class Device(PyverbsObject):
    """
    Device class represents the C ibv_device. It stores device's properties.
    It is not a part of objects creation order - there's no need for the user
    to create it for such purposes.
    """
    def __init__(self, name, guid, node_type, transport_type):
        self._node_type = node_type
        self._transport_type = transport_type
        self._name = name
        self._guid = guid

    @property
    def name(self):
        return self._name

    @property
    def node_type(self):
        return self._node_type

    @property
    def transport_type(self):
        return self._transport_type

    @property
    def guid(self):
        return self._guid

    def __str__(self):
        return 'Device {dev}, node type {ntype}, transport type {ttype},' \
               ' guid {guid}'.format(dev=self.name.decode(),
                ntype=translate_node_type(self.node_type),
                ttype=translate_transport_type(self.transport_type),
                guid=guid_to_hex(self.guid))


cdef class Context(PyverbsCM):
    """
    Context class represents the C ibv_context.
    """
    def __cinit__(self, **kwargs):
        """
        Initializes a Context object. The function searches the IB devices list
        for a device with the name provided by the user. If such a device is
        found, it is opened.
        :param kwargs: Currently supports 'name' argument only, the IB device's
                       name.
        :return: None
        """
        cdef int count
        cdef v.ibv_device **dev_list

        self.pds = weakref.WeakSet()
        self.dms = weakref.WeakSet()
        self.ccs = weakref.WeakSet()
        self.cqs = weakref.WeakSet()
        self.qps = weakref.WeakSet()

        dev_name = kwargs.get('name')

        if dev_name is not None:
            self.name = dev_name
        else:
            raise PyverbsUserError('Device name must be provided')

        dev_list = v.ibv_get_device_list(&count)
        if dev_list == NULL:
            raise PyverbsRDMAError('Failed to get devices list')
        try:
            for i in range(count):
                if dev_list[i].name.decode() == self.name:
                    self.context = v.ibv_open_device(dev_list[i])
                    if self.context == NULL:
                        raise PyverbsRDMAErrno('Failed to open device {dev}'.
                                               format(dev=self.name))
                    self.logger.debug('Context: opened device {dev}'.
                                      format(dev=self.name))
                    break
            else:
                raise PyverbsRDMAError('Failed to find device {dev}'.
                                       format(dev=self.name))
        finally:
            v.ibv_free_device_list(dev_list)

    def __dealloc__(self):
        """
        Closes the inner IB device.
        :return: None
        """
        self.close()

    cpdef close(self):
        self.logger.debug('Closing Context')
        self.close_weakrefs([self.qps, self.ccs, self.cqs, self.dms, self.pds])
        if self.context != NULL:
            rc = v.ibv_close_device(self.context)
            if rc != 0:
                raise PyverbsRDMAErrno('Failed to close device {dev}'.
                                       format(dev=self.device.name), errno)
            self.context = NULL

    @property
    def num_comp_vectors(self):
        return self.context.num_comp_vectors

    def query_device(self):
        """
        Queries the device's attributes.
        :return: A DeviceAttr object which holds the device's attributes as
                 reported by the hardware.
        """
        dev_attr = DeviceAttr()
        rc = v.ibv_query_device(self.context, &dev_attr.dev_attr)
        if rc != 0:
            raise PyverbsRDMAErrno('Failed to query device {name}'.
                                   format(name=self.name), errno)
        return dev_attr

    def query_device_ex(self, QueryDeviceExInput ex_input = None):
        """
        Queries the device's extended attributes.
        :param ex_input: An extensible input struct for possible future
                         extensions
        :return: DeviceAttrEx object
        """
        dev_attr_ex = DeviceAttrEx()
        rc = v.ibv_query_device_ex(self.context,
                                   &ex_input.input if ex_input is not None else NULL,
                                   &dev_attr_ex.dev_attr)
        if rc != 0:
            raise PyverbsRDMAErrno('Failed to query EX device {name}'.
                                   format(name=self.name))
        return dev_attr_ex

    def query_gid(self, unsigned int port_num, int index):
        gid = GID()
        rc = v.ibv_query_gid(self.context, port_num, index, &gid.gid)
        if rc != 0:
            raise PyverbsRDMAError('Failed to query gid {idx} of port {port}'.
                                                                   format(idx=index, port=port_num))
        return gid

    def query_port(self, unsigned int port_num):
        """
        Query port <port_num> of the device and returns its attributes.
        :param port_num: Port number to query
        :return: PortAttr object on success
        """
        port_attrs = PortAttr()
        rc = v.ibv_query_port(self.context, port_num, &port_attrs.attr)
        if rc != 0:
            raise PyverbsRDMAErrno('Failed to query port {p}'.format(p=port_num))
        return port_attrs

    cdef add_ref(self, obj):
        if isinstance(obj, PD):
            self.pds.add(obj)
        elif isinstance(obj, DM):
            self.dms.add(obj)
        elif isinstance(obj, CompChannel):
            self.ccs.add(obj)
        elif isinstance(obj, CQ) or isinstance(obj, CQEX):
            self.cqs.add(obj)
        elif isinstance(obj, QP):
            self.qps.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')


cdef class DeviceAttr(PyverbsObject):
    """
    DeviceAttr represents ibv_device_attr C class. It exposes the same
    properties (read only) and also provides an __str__() function for
    readability.
    """
    @property
    def fw_version(self):
        return self.dev_attr.fw_ver.decode()
    @property
    def node_guid(self):
        return self.dev_attr.node_guid
    @property
    def sys_image_guid(self):
        return self.dev_attr.sys_image_guid
    @property
    def max_mr_size(self):
        return self.dev_attr.max_mr_size
    @property
    def page_size_cap(self):
        return self.dev_attr.page_size_cap
    @property
    def vendor_id(self):
        return self.dev_attr.vendor_id
    @property
    def vendor_part_id(self):
        return self.dev_attr.vendor_part_id
    @property
    def hw_ver(self):
        return self.dev_attr.hw_ver
    @property
    def max_qp(self):
        return self.dev_attr.max_qp
    @property
    def max_qp_wr(self):
        return self.dev_attr.max_qp_wr
    @property
    def device_cap_flags(self):
        return self.dev_attr.device_cap_flags
    @property
    def max_sge(self):
        return self.dev_attr.max_sge
    @property
    def max_sge_rd(self):
        return self.dev_attr.max_sge_rd
    @property
    def max_cq(self):
        return self.dev_attr.max_cq
    @property
    def max_cqe(self):
        return self.dev_attr.max_cqe
    @property
    def max_mr(self):
        return self.dev_attr.max_mr
    @property
    def max_pd(self):
        return self.dev_attr.max_pd
    @property
    def max_qp_rd_atom(self):
        return self.dev_attr.max_qp_rd_atom
    @property
    def max_ee_rd_atom(self):
        return self.dev_attr.max_ee_rd_atom
    @property
    def max_res_rd_atom(self):
        return self.dev_attr.max_res_rd_atom
    @property
    def max_qp_init_rd_atom(self):
        return self.dev_attr.max_qp_init_rd_atom
    @property
    def max_ee_init_rd_atom(self):
        return self.dev_attr.max_ee_init_rd_atom
    @property
    def atomic_caps(self):
        return self.dev_attr.atomic_cap
    @property
    def max_ee(self):
        return self.dev_attr.max_ee
    @property
    def max_rdd(self):
        return self.dev_attr.max_rdd
    @property
    def max_mw(self):
        return self.dev_attr.max_mw
    @property
    def max_raw_ipv6_qps(self):
        return self.dev_attr.max_raw_ipv6_qp
    @property
    def max_raw_ethy_qp(self):
        return self.dev_attr.max_raw_ethy_qp
    @property
    def max_mcast_grp(self):
        return self.dev_attr.max_mcast_grp
    @property
    def max_mcast_qp_attach(self):
        return self.dev_attr.max_mcast_qp_attach
    @property
    def max_ah(self):
        return self.dev_attr.max_ah
    @property
    def max_fmr(self):
        return self.dev_attr.max_fmr
    @property
    def max_map_per_fmr(self):
        return self.dev_attr.max_map_per_fmr
    @property
    def max_srq(self):
        return self.dev_attr.max_srq
    @property
    def max_srq_wr(self):
        return self.dev_attr.max_srq_wr
    @property
    def max_srq_sge(self):
        return self.dev_attr.max_srq_sge
    @property
    def max_pkeys(self):
        return self.dev_attr.max_pkeys
    @property
    def local_ca_ack_delay(self):
        return self.dev_attr.local_ca_ack_delay
    @property
    def phys_port_cnt(self):
        return self.dev_attr.phys_port_cnt

    def __str__(self):
        print_format = '{:<22}: {:<20}\n'
        return print_format.format('FW version', self.fw_version) +\
            print_format.format('Node guid', guid_format(self.node_guid)) +\
            print_format.format('Sys image GUID', guid_format(self.sys_image_guid)) +\
            print_format.format('Max MR size', hex(self.max_mr_size).replace('L', '')) +\
            print_format.format('Page size cap', hex(self.page_size_cap).replace('L', '')) +\
            print_format.format('Vendor ID', hex(self.vendor_id)) +\
            print_format.format('Vendor part ID', self.vendor_part_id) +\
            print_format.format('HW version', self.hw_ver) +\
            print_format.format('Max QP', self.max_qp) +\
            print_format.format('Max QP WR', self.max_qp_wr) +\
            print_format.format('Device cap flags',
                                translate_device_caps(self.device_cap_flags)) +\
            print_format.format('Max SGE', self.max_sge) +\
            print_format.format('Max SGE RD', self.max_sge_rd) +\
            print_format.format('MAX CQ', self.max_cq) +\
            print_format.format('Max CQE', self.max_cqe) +\
            print_format.format('Max MR', self.max_mr) +\
            print_format.format('Max PD', self.max_pd) +\
            print_format.format('Max QP RD atom', self.max_qp_rd_atom) +\
            print_format.format('Max EE RD atom', self.max_ee_rd_atom) +\
            print_format.format('Max res RD atom', self.max_res_rd_atom) +\
            print_format.format('Max QP init RD atom', self.max_qp_init_rd_atom) +\
            print_format.format('Max EE init RD atom', self.max_ee_init_rd_atom) +\
            print_format.format('Atomic caps', self.atomic_caps) +\
            print_format.format('Max EE', self.max_ee) +\
            print_format.format('Max RDD', self.max_rdd) +\
            print_format.format('Max MW', self.max_mw) +\
            print_format.format('Max raw IPv6 QPs', self.max_raw_ipv6_qps) +\
            print_format.format('Max raw ethy QP', self.max_raw_ethy_qp) +\
            print_format.format('Max mcast group', self.max_mcast_grp) +\
            print_format.format('Max mcast QP attach', self.max_mcast_qp_attach) +\
            print_format.format('Max AH', self.max_ah) +\
            print_format.format('Max FMR', self.max_fmr) +\
            print_format.format('Max map per FMR', self.max_map_per_fmr) +\
            print_format.format('Max SRQ', self.max_srq) +\
            print_format.format('Max SRQ WR', self.max_srq_wr) +\
            print_format.format('Max SRQ SGE', self.max_srq_sge) +\
            print_format.format('Max PKeys', self.max_pkeys) +\
            print_format.format('local CA ack delay', self.local_ca_ack_delay) +\
            print_format.format('Phys port count', self.phys_port_cnt)


cdef class QueryDeviceExInput(PyverbsObject):
    def __cinit__(self, comp_mask):
        self.ex_input.comp_mask = comp_mask


cdef class ODPCaps(PyverbsObject):
    @property
    def general_caps(self):
        return self.odp_caps.general_caps
    @property
    def rc_odp_caps(self):
        return self.odp_caps.per_transport_caps.rc_odp_caps
    @property
    def uc_odp_caps(self):
        return self.odp_caps.per_transport_caps.uc_odp_caps
    @property
    def ud_odp_caps(self):
        return self.odp_caps.per_transport_caps.ud_odp_caps


cdef class TSOCaps(PyverbsObject):
    @property
    def max_tso(self):
        return self.tso_caps.max_tso
    @property
    def supported_qpts(self):
        return self.tso_caps.supported_qpts


cdef class RSSCaps(PyverbsObject):
    @property
    def supported_qpts(self):
        return self.rss_caps.supported_qpts
    @property
    def max_rwq_indirection_tables(self):
        return self.rss_caps.max_rwq_indirection_tables
    @property
    def rx_hash_fields_mask(self):
        return self.rss_caps.rx_hash_fields_mask
    @property
    def rx_hash_function(self):
        return self.rss_caps.rx_hash_function
    @property
    def max_rwq_indirection_table_size(self):
        return self.rss_caps.max_rwq_indirection_table_size


cdef class PacketPacingCaps(PyverbsObject):
    @property
    def qp_rate_limit_min(self):
        return self.packet_pacing_caps.qp_rate_limit_min
    @property
    def qp_rate_limit_max(self):
        return self.packet_pacing_caps.qp_rate_limit_max
    @property
    def supported_qpts(self):
        return self.packet_pacing_caps.supported_qpts


cdef class TMCaps(PyverbsObject):
    @property
    def max_rndv_hdr_size(self):
        return self.tm_caps.max_rndv_hdr_size
    @property
    def max_num_tags(self):
        return self.tm_caps.max_num_tags
    @property
    def flags(self):
        return self.tm_caps.flags
    @property
    def max_ops(self):
        return self.tm_caps.max_ops
    @property
    def max_sge(self):
        return self.tm_caps.max_sge


cdef class CQModerationCaps(PyverbsObject):
    @property
    def max_cq_count(self):
        return self.cq_mod_caps.max_cq_count
    @property
    def max_cq_period(self):
        return self.cq_mod_caps.max_cq_period


cdef class DeviceAttrEx(PyverbsObject):
    @property
    def orig_attr(self):
        attr = DeviceAttr()
        attr.dev_attr = self.dev_attr.orig_attr
        return attr
    @property
    def comp_mask(self):
        return self.dev_attr.comp_mask
    @comp_mask.setter
    def comp_mask(self, val):
        self.dev_attr.comp_mask = val
    @property
    def odp_caps(self):
        caps = ODPCaps()
        caps.odp_caps = self.dev_attr.odp_caps
        return caps
    @property
    def completion_timestamp_mask(self):
        return self.dev_attr.completion_timestamp_mask
    @property
    def hca_core_clock(self):
        return self.dev_attr.hca_core_clock
    @property
    def device_cap_flags_ex(self):
        return self.dev_attr.device_cap_flags_ex
    @property
    def tso_caps(self):
        caps = TSOCaps()
        caps.tso_caps = self.dev_attr.tso_caps
        return caps
    @property
    def rss_caps(self):
        caps = RSSCaps()
        caps.rss_caps = self.dev_attr.rss_caps
        return caps
    @property
    def max_wq_type_rq(self):
        return self.dev_attr.max_wq_type_rq
    @property
    def packet_pacing_caps(self):
        caps = PacketPacingCaps()
        caps.packet_pacing_caps = self.dev_attr.packet_pacing_caps
        return caps
    @property
    def raw_packet_caps(self):
        return self.dev_attr.raw_packet_caps
    @property
    def tm_caps(self):
        caps = TMCaps()
        caps.tm_caps = self.dev_attr.tm_caps
        return caps
    @property
    def cq_mod_caps(self):
        caps = CQModerationCaps()
        caps.cq_mod_caps = self.dev_attr.cq_mod_caps
        return caps
    @property
    def max_dm_size(self):
        return self.dev_attr.max_dm_size


cdef class AllocDmAttr(PyverbsObject):
    def __cinit__(self, length, log_align_req = 0, comp_mask = 0):
        """
        Creates an AllocDmAttr object with the given parameters. This object
        can than be used to create a DM object.
        :param length: Length of the future device memory
        :param log_align_req: log2 of address alignment requirement
        :param comp_mask: compatibility mask
        :return: An AllocDmAttr object
        """
        self.alloc_dm_attr.length = length
        self.alloc_dm_attr.log_align_req = log_align_req
        self.alloc_dm_attr.comp_mask = comp_mask

    @property
    def length(self):
        return self.alloc_dm_attr.length

    @length.setter
    def length(self, val):
        self.alloc_dm_attr.length = val

    @property
    def log_align_req(self):
        return self.alloc_dm_attr.log_align_req

    @log_align_req.setter
    def log_align_req(self, val):
        self.alloc_dm_attr.log_align_req = val

    @property
    def comp_mask(self):
        return self.alloc_dm_attr.comp_mask

    @comp_mask.setter
    def comp_mask(self, val):
        self.alloc_dm_attr.comp_mask = val


cdef class DM(PyverbsCM):
    def __cinit__(self, Context context, AllocDmAttr dm_attr not None):
        """
        Allocate a device (direct) memory.
        :param context: The context of the device on which to allocate memory
        :param dm_attr: Attributes that define the DM
        :return: A DM object on success
        """
        self.dm_mrs = weakref.WeakSet()
        device_attr = context.query_device_ex()
        if device_attr.max_dm_size <= 0:
            raise PyverbsUserError('Device doesn\'t support dm allocation')
        self.dm = v.ibv_alloc_dm(<v.ibv_context*>context.context,
                                 &dm_attr.alloc_dm_attr)
        if self.dm == NULL:
            raise PyverbsRDMAErrno('Failed to allocate device memory of size '
                                   '{size}. Max available size {max}.'
                                   .format(size=dm_attr.length,
                                           max=device_attr.max_dm_size))
        self.context = context
        context.add_ref(self)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        self.logger.debug('Closing DM')
        self.close_weakrefs([self.dm_mrs])
        if self.dm != NULL:
            rc = v.ibv_free_dm(self.dm)
            if rc != 0:
                raise PyverbsRDMAErrno('Failed to free dm')
            self.dm = NULL
        self.context = None

    cdef add_ref(self, obj):
        if isinstance(obj, DMMR):
            self.dm_mrs.add(obj)

    def copy_to_dm(self, dm_offset, data, length):
        rc = v.ibv_memcpy_to_dm(<v.ibv_dm *>self.dm, <uint64_t>dm_offset,
                                <char *>data, <size_t>length)
        if rc != 0:
            raise PyverbsRDMAErrno('Failed to copy to dm')

    def copy_from_dm(self, dm_offset, length):
        cdef char *data =<char*>malloc(length)
        memset(data, 0, length)
        rc = v.ibv_memcpy_from_dm(<void *>data, <v.ibv_dm *>self.dm,
                                  <uint64_t>dm_offset, <size_t>length)
        if rc != 0:
            raise PyverbsRDMAErrno('Failed to copy from dm')
        res = data[:length]
        free(data)
        return res


cdef class PortAttr(PyverbsObject):
    @property
    def state(self):
        return self.attr.state
    @property
    def max_mtu(self):
        return self.attr.max_mtu
    @property
    def active_mtu(self):
        return self.attr.active_mtu
    @property
    def gid_tbl_len(self):
        return self.attr.gid_tbl_len
    @property
    def port_cap_flags(self):
        return self.attr.port_cap_flags
    @property
    def max_msg_sz(self):
        return self.attr.max_msg_sz
    @property
    def bad_pkey_cntr(self):
        return self.attr.bad_pkey_cntr
    @property
    def qkey_viol_cntr(self):
        return self.attr.qkey_viol_cntr
    @property
    def pkey_tbl_len(self):
        return self.attr.pkey_tbl_len
    @property
    def lid(self):
        return self.attr.lid
    @property
    def sm_lid(self):
        return self.attr.sm_lid
    @property
    def lmc(self):
        return self.attr.lmc
    @property
    def max_vl_num(self):
        return self.attr.max_vl_num
    @property
    def sm_sl(self):
        return self.attr.sm_sl
    @property
    def subnet_timeout(self):
        return self.attr.subnet_timeout
    @property
    def init_type_reply(self):
        return self.attr.init_type_reply
    @property
    def active_width(self):
        return self.attr.active_width
    @property
    def active_speed(self):
        return self.attr.active_speed
    @property
    def phys_state(self):
        return self.attr.phys_state
    @property
    def link_layer(self):
        return self.attr.link_layer
    @property
    def flags(self):
        return self.attr.flags
    @property
    def port_cap_flags2(self):
        return self.attr.port_cap_flags2

    def __str__(self):
        print_format = '{:<24}: {:<20}\n'
        return print_format.format('Port state', port_state_to_str(self.attr.state)) +\
            print_format.format('Max MTU', translate_mtu(self.attr.max_mtu)) +\
            print_format.format('Active MTU', translate_mtu(self.attr.active_mtu)) +\
            print_format.format('SM lid', self.attr.sm_lid) +\
            print_format.format('Port lid', self.attr.lid) +\
            print_format.format('lmc', hex(self.attr.lmc)) +\
            print_format.format('Link layer', translate_link_layer(self.attr.link_layer)) +\
            print_format.format('Max message size', hex(self.attr.max_msg_sz)) +\
            print_format.format('Port cap flags', translate_port_cap_flags(self.attr.port_cap_flags)) +\
            print_format.format('Port cap flags 2', translate_port_cap_flags2(self.attr.port_cap_flags2)) +\
            print_format.format('max VL num', self.attr.max_vl_num) +\
            print_format.format('Bad Pkey counter', self.attr.bad_pkey_cntr) +\
            print_format.format('Qkey violations counter', self.attr.qkey_viol_cntr) +\
            print_format.format('GID table len', self.attr.gid_tbl_len) +\
            print_format.format('Pkey table len', self.attr.pkey_tbl_len) +\
            print_format.format('SM sl', self.attr.sm_sl) +\
            print_format.format('Subnet timeout', self.attr.subnet_timeout) +\
            print_format.format('Init type reply', self.attr.init_type_reply) +\
            print_format.format('Active width', width_to_str(self.attr.active_width)) +\
            print_format.format('Active speed', speed_to_str(self.attr.active_speed)) +\
            print_format.format('Phys state', phys_state_to_str(self.attr.phys_state)) +\
            print_format.format('Flags', self.attr.flags)


def guid_format(num):
    """
    Get GUID representation of the given number, including change of endianness.
    :param num: Number to change to GUID format.
    :return: GUID-formatted string.
    """
    num = be64toh(num)
    hex_str = "%016x" % (num)
    hex_array = [hex_str[i:i+2] for i in range(0, len(hex_str), 2)]
    hex_array = [''.join(x) for x in zip(hex_array[0::2], hex_array[1::2])]
    return ':'.join(hex_array)


def translate_transport_type(transport_type):
    l = {0: 'IB', 1: 'IWARP', 2: 'USNIC', 3: 'USNIC UDP'}
    try:
        return l[transport_type]
    except KeyError:
        return 'Unknown'


def translate_node_type(node_type):
    l = {1: 'CA', 2: 'Switch', 3: 'Router', 4: 'RNIC', 5: 'USNIC',
         6: 'USNIC UDP'}
    try:
        return l[node_type]
    except KeyError:
        return 'Unknown'


def guid_to_hex(node_guid):
    return hex(node_guid).replace('L', '').replace('0x', '')


def port_state_to_str(port_state):
    l = {0: 'NOP', 1: 'Down', 2: 'Init', 3: 'Armed', 4: 'Active', 5: 'Defer'}
    try:
        return '{s} ({n})'.format(s=l[port_state].name, n=port_state)
    except KeyError:
        return 'Invalid state ({s})'.format(s=port_state)


def translate_mtu(mtu):
    l = {1: 256, 2: 512, 3: 1024, 4: 2048, 5: 4096}
    try:
        return '{s} ({n})'.format(s=l[mtu], n=mtu)
    except KeyError:
        return 'Invalid MTU ({m})'.format(m=mtu)


def translate_link_layer(ll):
    l = {0: 'Unspecified', 1:'InfiniBand', 2:'Ethernet'}
    try:
        return l[ll]
    except KeyError:
        return 'Invalid link layer ({ll})'.format(ll=ll)


def translate_port_cap_flags(flags):
    l = {e.IBV_PORT_SM: 'IBV_PORT_SM',
         e.IBV_PORT_NOTICE_SUP: 'IBV_PORT_NOTICE_SUP',
         e.IBV_PORT_TRAP_SUP: 'IBV_PORT_TRAP_SUP',
         e.IBV_PORT_OPT_IPD_SUP: 'IBV_PORT_OPT_IPD_SUP',
         e.IBV_PORT_AUTO_MIGR_SUP: 'IBV_PORT_AUTO_MIGR_SUP',
         e.IBV_PORT_SL_MAP_SUP: 'IBV_PORT_SL_MAP_SUP',
         e.IBV_PORT_MKEY_NVRAM: 'IBV_PORT_MKEY_NVRAM',
         e.IBV_PORT_PKEY_NVRAM: 'IBV_PORT_PKEY_NVRAM',
         e.IBV_PORT_LED_INFO_SUP: 'IBV_PORT_LED_INFO_SUP',
         e.IBV_PORT_SYS_IMAGE_GUID_SUP: 'IBV_PORT_SYS_IMAGE_GUID_SUP',
         e.IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP: 'IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP',
         e.IBV_PORT_EXTENDED_SPEEDS_SUP: 'IBV_PORT_EXTENDED_SPEEDS_SUP',
         e.IBV_PORT_CAP_MASK2_SUP: 'IBV_PORT_CAP_MASK2_SUP',
         e.IBV_PORT_CM_SUP: 'IBV_PORT_CM_SUP',
         e.IBV_PORT_SNMP_TUNNEL_SUP: 'IBV_PORT_SNMP_TUNNEL_SUP',
         e.IBV_PORT_REINIT_SUP: 'IBV_PORT_REINIT_SUP',
         e.IBV_PORT_DEVICE_MGMT_SUP: 'IBV_PORT_DEVICE_MGMT_SUP',
         e.IBV_PORT_VENDOR_CLASS_SUP: 'IBV_PORT_VENDOR_CLASS_SUP',
         e.IBV_PORT_DR_NOTICE_SUP: 'IBV_PORT_DR_NOTICE_SUP',
         e.IBV_PORT_CAP_MASK_NOTICE_SUP: 'IBV_PORT_CAP_MASK_NOTICE_SUP',
         e.IBV_PORT_BOOT_MGMT_SUP: 'IBV_PORT_BOOT_MGMT_SUP',
         e.IBV_PORT_LINK_LATENCY_SUP: 'IBV_PORT_LINK_LATENCY_SUP',
         e.IBV_PORT_CLIENT_REG_SUP: 'IBV_PORT_CLIENT_REG_SUP',
         e.IBV_PORT_IP_BASED_GIDS: 'IBV_PORT_IP_BASED_GIDS'}
    return str_from_flags(flags, l)


def translate_port_cap_flags2(flags):
    l = {e.IBV_PORT_SET_NODE_DESC_SUP: 'IBV_PORT_SET_NODE_DESC_SUP',
         e.IBV_PORT_INFO_EXT_SUP: 'IBV_PORT_INFO_EXT_SUP',
         e.IBV_PORT_VIRT_SUP: 'IBV_PORT_VIRT_SUP',
         e.IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP: 'IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP',
         e.IBV_PORT_LINK_WIDTH_2X_SUP: 'IBV_PORT_LINK_WIDTH_2X_SUP',
         e.IBV_PORT_LINK_SPEED_HDR_SUP: 'IBV_PORT_LINK_SPEED_HDR_SUP'}
    return str_from_flags(flags, l)


def translate_device_caps(flags):
    l = {e.IBV_DEVICE_RESIZE_MAX_WR: 'IBV_DEVICE_RESIZE_MAX_WR',
         e.IBV_DEVICE_BAD_PKEY_CNTR: 'IBV_DEVICE_BAD_PKEY_CNTR',
         e.IBV_DEVICE_BAD_QKEY_CNTR: 'IBV_DEVICE_BAD_QKEY_CNTR',
         e.IBV_DEVICE_RAW_MULTI: 'IBV_DEVICE_RAW_MULTI',
         e.IBV_DEVICE_AUTO_PATH_MIG: 'IBV_DEVICE_AUTO_PATH_MIG',
         e.IBV_DEVICE_CHANGE_PHY_PORT: 'IBV_DEVICE_CHANGE_PHY_PORT',
         e.IBV_DEVICE_UD_AV_PORT_ENFORCE: 'IBV_DEVICE_UD_AV_PORT_ENFORCE',
         e.IBV_DEVICE_CURR_QP_STATE_MOD: 'IBV_DEVICE_CURR_QP_STATE_MOD',
         e.IBV_DEVICE_SHUTDOWN_PORT: 'IBV_DEVICE_SHUTDOWN_PORT',
         e.IBV_DEVICE_INIT_TYPE: 'IBV_DEVICE_INIT_TYPE',
         e.IBV_DEVICE_PORT_ACTIVE_EVENT: 'IBV_DEVICE_PORT_ACTIVE_EVENT',
         e.IBV_DEVICE_SYS_IMAGE_GUID: 'IBV_DEVICE_SYS_IMAGE_GUID',
         e.IBV_DEVICE_RC_RNR_NAK_GEN: 'IBV_DEVICE_RC_RNR_NAK_GEN',
         e.IBV_DEVICE_SRQ_RESIZE: 'IBV_DEVICE_SRQ_RESIZE',
         e.IBV_DEVICE_N_NOTIFY_CQ: 'IBV_DEVICE_N_NOTIFY_CQ',
         e.IBV_DEVICE_MEM_WINDOW: 'IBV_DEVICE_MEM_WINDOW',
         e.IBV_DEVICE_UD_IP_CSUM: 'IBV_DEVICE_UD_IP_CSUM',
         e.IBV_DEVICE_XRC: 'IBV_DEVICE_XRC',
         e.IBV_DEVICE_MEM_MGT_EXTENSIONS: 'IBV_DEVICE_MEM_MGT_EXTENSIONS',
         e.IBV_DEVICE_MEM_WINDOW_TYPE_2A: 'IBV_DEVICE_MEM_WINDOW_TYPE_2A',
         e.IBV_DEVICE_MEM_WINDOW_TYPE_2B: 'IBV_DEVICE_MEM_WINDOW_TYPE_2B',
         e.IBV_DEVICE_RC_IP_CSUM: 'IBV_DEVICE_RC_IP_CSUM',
         e.IBV_DEVICE_RAW_IP_CSUM: 'IBV_DEVICE_RAW_IP_CSUM',
         e.IBV_DEVICE_MANAGED_FLOW_STEERING: 'IBV_DEVICE_MANAGED_FLOW_STEERING'}
    return str_from_flags(flags, l)


def str_from_flags(flags, dictionary):
    str_flags = "\n  "
    for bit in dictionary:
        if flags & bit:
            str_flags += dictionary[bit]
            str_flags += '\n  '
    return str_flags


def phys_state_to_str(phys):
    l =  {1: 'Sleep', 2: 'Polling', 3: 'Disabled',
          4: 'Port configuration training', 5: 'Link up',
          6: 'Link error recovery', 7: 'Phy test'}
    try:
        return '{s} ({n})'.format(s=l[phys], n=phys)
    except KeyError:
        return 'Invalid physical state'


def width_to_str(width):
    l = {1: '1X', 2: '4X', 4: '8X', 16: '2X'}
    try:
        return '{s} ({n})'.format(s=l[width], n=width)
    except KeyError:
        return 'Invalid width'


def speed_to_str(speed):
    l = {1: '2.5 Gbps', 2: '5.0 Gbps', 4: '5.0 Gbps', 8: '10.0 Gbps',
         16: '14.0 Gbps', 32: '25.0 Gbps', 64: '50.0 Gbps'}
    try:
        return '{s} ({n})'.format(s=l[speed], n=speed)
    except KeyError:
        return 'Invalid speed'


def get_device_list():
    """
    :return: list of IB_devices on current node
             each list element contains a Device with:
                 device name
                 device node type
                 device transport type
                 device guid
    """
    cdef int count = 0;
    cdef v.ibv_device **dev_list;
    dev_list = v.ibv_get_device_list(&count)
    if dev_list == NULL:
            raise PyverbsRDMAError('Failed to get devices list')
    devices = []
    try:
        for i in range(count):
            name = dev_list[i].name
            node = dev_list[i].node_type
            transport = dev_list[i].transport_type
            guid = be64toh(v.ibv_get_device_guid(dev_list[i]))
            devices.append(Device(name, guid, node, transport))
    finally:
        v.ibv_free_device_list(dev_list)
    return devices



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved.  See COPYING file

from pyverbs.addr import GlobalRoute, AHAttr, AH
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.tests.base import PyverbsTestCase
import pyverbs.device as d
import pyverbs.enums as e
from pyverbs.pd import PD
from pyverbs.cq import WC

class AHTest(PyverbsTestCase):
    """
    Test various functionalities of the AH class.
    """
    def test_create_ah(self):
        """
        Test ibv_create_ah.
        """
        for ctx, attr, attr_ex in self.devices:
            pd = PD(ctx)
            for port_num in range(1, 1 + attr.phys_port_cnt):
                gr = get_global_route(ctx, port_num=port_num)
                ah_attr = AHAttr(gr=gr, is_global=1, port_num=port_num)
                with AH(pd, attr=ah_attr):
                    pass
    # TODO: Test ibv_create_ah_from_wc once we have traffic

    def test_create_ah_roce(self):
        """
        Verify that AH can't be created without GRH in RoCE
        """
        for ctx, attr, attr_ex in self.devices:
            pd = PD(ctx)
            for port_num in range(1, 1 + attr.phys_port_cnt):
                port_attr = ctx.query_port(port_num)
                if port_attr.link_layer == e.IBV_LINK_LAYER_INFINIBAND:
                    return
                ah_attr = AHAttr(is_global=0, port_num=port_num)
                try:
                    ah = AH(pd, attr=ah_attr)
                except PyverbsError as err:
                    assert 'Failed to create AH' in err.args[0]
                else:
                    raise PyverbsError('Created a non-global AH on RoCE')

    def test_destroy_ah(self):
        """
        Test ibv_destroy_ah.
        """
        for ctx, attr, attr_ex in self.devices:
            pd = PD(ctx)
            for port_num in range(1, 1 + attr.phys_port_cnt):
                gr = get_global_route(ctx)
                ah_attr = AHAttr(gr=gr, is_global=1, port_num=port_num)
                with AH(pd, attr=ah_attr) as ah:
                    ah.close()


def get_global_route(ctx, gid_index=0, port_num=1):
    """
    Queries the provided Context's gid <gid_index> and creates a GlobalRoute
    object with sgid_index <gid_index> and the queried GID as dgid.
    :param ctx: Context object to query
    :param gid_index: GID index to query and use. Default: 0, as it's always
                      valid
    :param port_num: Number of the port to query. Default: 1
    :return: GlobalRoute object
    """
    gid = ctx.query_gid(port_num, gid_index)
    gr = GlobalRoute(dgid=gid, sgid_index=gid_index)
    return gr



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019, Mellanox Technologies. All rights reserved. See COPYING file

from pyverbs.pyverbs_error import PyverbsRDMAError, PyverbsError
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.device cimport DM
from .pd cimport PD
import resource

cdef extern from 'stdlib.h':
    int posix_memalign(void **memptr, size_t alignment, size_t size)
cdef extern from 'stdlib.h':
    void free(void *ptr)
cdef extern from 'string.h':
    void *memcpy(void *dest, const void *src, size_t n)
    void *memset(void *s, int c, size_t n)
cdef extern from 'stdint.h':
    ctypedef int uintptr_t


cdef class MR(PyverbsCM):
    """
    MR class represents ibv_mr. Buffer allocation in done in the c'tor. Freeing
    it is done in close().
    """
    def __cinit__(self, PD pd not None, length, access, **kwargs):
        """
        Allocate a user-level buffer of length <length> and register a Memory
        Region of the given length and access flags.
        :param pd: A PD object
        :param length: Length in bytes
        :param access: Access flags, see ibv_access_flags enum
        :return: The newly created MR on success
        """
        if len(kwargs) != 0:
            return
        #We want to enable registering an MR of size 0 but this fails with a
        #buffer of size 0, so in this case lets increase the buffer
        if length == 0:
            length = 10
        rc = posix_memalign(&self.buf, resource.getpagesize(), length)
        if rc:
            raise PyverbsRDMAError('Failed to allocate MR buffer of size {l}'.
                                   format(l=length))
        memset(self.buf, 0, length)
        self.mr = v.ibv_reg_mr(<v.ibv_pd*>pd.pd, self.buf, length, access)
        if self.mr == NULL:
            raise PyverbsRDMAErrno('Failed to register a MR. length: {l}, access flags: {a}'.
                                   format(l=length, a=access))
        self.pd = pd
        pd.add_ref(self)
        self.logger.debug('Registered ibv_mr. Length: {l}, access flags {a}'.
                          format(l=length, a=access))

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        """
        Closes the underlying C object of the MR and frees the memory allocated.
        MR may be deleted directly or indirectly by closing its context, which
        leaves the Python PD object without the underlying C object, so during
        destruction, need to check whether or not the C object exists.
        :return: None
        """
        self.logger.debug('Closing MR')
        if self.mr != NULL:
            rc = v.ibv_dereg_mr(self.mr)
            if rc != 0:
                raise PyverbsRDMAErrno('Failed to dereg MR')
            self.mr = NULL
            self.pd = None
        free(self.buf)
        self.buf = NULL

    def write(self, data, length):
        """
        Write user data to the MR's buffer using memcpy
        :param data: User data to write
        :param length: Length of the data to write
        :return: None
        """
        # If data is a string, cast it to bytes as Python3 doesn't
        # automatically convert it.
        if isinstance(data, str):
            data = data.encode()
        memcpy(self.buf, <char *>data, length)

    cpdef read(self, length, offset):
        """
        Reads data from the MR's buffer
        :param length: Length of data to read
        :param offset: Reading offset
        :return: The data on the buffer in the requested offset
        """
        cdef char *data
        cdef int off = offset # we can't use offset in the next line, as it is
                              # a Python object and not C
        data = <char*>(self.buf + off)
        return data[:length]

    @property
    def buf(self):
        return <uintptr_t>self.buf

    @property
    def lkey(self):
        return self.mr.lkey

    @property
    def rkey(self):
        return self.mr.rkey


cdef class MW(PyverbsCM):
    def __cinit__(self, PD pd not None, v.ibv_mw_type mw_type):
        """
        Initializes a memory window object of the given type
        :param pd: A PD object
        :param mw_type: Type of of the memory window, see ibv_mw_type enum
        :return:
        """
        self.mw = NULL
        self.mw = v.ibv_alloc_mw(pd.pd, mw_type)
        if self.mw == NULL:
            raise PyverbsRDMAErrno('Failed to allocate MW')
        self.pd = pd
        pd.add_ref(self)
        self.logger.debug('Allocated memory window of type {t}'.
                          format(t=mwtype2str(mw_type)))

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        """
        Closes the underlaying C MW object.
        MW may be deleted directly or by deleting its PD, which leaves the
        Python object without the underlaying MW.
        Need to check that the underlaying MW wasn't dealloced before.
        :return: None
        """
        self.logger.debug('Closing MW')
        if self.mw is not NULL:
            rc = v.ibv_dealloc_mw(self.mw)
            if rc != 0:
               raise PyverbsRDMAErrno('Failed to dealloc MW')
            self.mw = NULL
            self.pd = None


cdef class DMMR(MR):
    def __cinit__(self, PD pd not None, length, access, **kwargs):
        """
        Initializes a DMMR (Device Memory Memory Region) of the given length
        and access flags using the given PD and DM objects.
        :param pd: A PD object
        :param length: Length in bytes
        :param access: Access flags, see ibv_access_flags enum
        :param kwargs: see below
        :return: The newly create DMMR

        :keyword Arguments:
            * *dm* (DM)
               A DM (device memory) object to be used for this DMMR
            * *offset*
               Byte offset from the beginning of the allocated device memory
               buffer
        """
        dm = <DM>kwargs['dm']
        offset = kwargs['offset']
        self.mr = v.ibv_reg_dm_mr(pd.pd, (<DM>dm).dm, offset, length, access)
        if self.mr == NULL:
            raise PyverbsRDMAErrno('Failed to register a device MR. length: {len}, access flags: {flags}'.
                                   format(len=length, flags=access,))
        self.pd = pd
        self.dm = dm
        pd.add_ref(self)
        dm.add_ref(self)
        self.logger.debug('Registered device ibv_mr. Length: {len}, access flags {flags}'.
                          format(len=length, flags=access))

    def write(self, data, length):
        return self.dm.copy_to_dm(0, data, length)

    cpdef read(self, length, offset):
        return self.dm.copy_from_dm(offset, length)


def mwtype2str(mw_type):
    mw_types = {1:'IBV_MW_TYPE_1', 2:'IBV_MW_TYPE_2'}
    try:
        return mw_types[mw_type]
    except KeyError:
        return 'Unknown MW type ({t})'.format(t=mw_type)



import re
import sys

def global_ln(ln):
    g = re.match(r"^#define\s+(\S+)\s+CL_HTON(\d+)\((.*)\)",ln)
    if g:
        print("#define %s htobe%s(%s)"%(g.group(1),g.group(2),g.group(3)))
        return global_ln
    g = re.match(r"^#define\s+(\S+)\s+\(CL_HTON(\d+)\((.*)\)\)",ln)
    if g:
        print("#define %s htobe%s(%s)"%(g.group(1),g.group(2),g.group(3)))
        return global_ln
    g = re.match(r"^#define\s+(\S+)\s+(0x\w+)",ln)
    if g:
        print("#define %s %s"%(g.group(1),g.group(2)))
        return global_ln
    g = re.match(r"^#define\s+(\S+)\s+\((0x\w+)\)",ln)
    if g:
        print("#define %s %s"%(g.group(1),g.group(2)))
        return global_ln
    g = re.match(r"^#define\s+(\S+)\s+(\d+)",ln)
    if g:
        print("#define %s %s"%(g.group(1),g.group(2)))
        return global_ln
    g = re.match(r"^#define\s+(\S+)\s+\((\d+)\)",ln)
    if g:
        print("#define %s %s"%(g.group(1),g.group(2)))
        return global_ln

    g = re.match(r"^typedef\s+(union|struct)\s+_\S+\s+{",ln);
    if g:
        print("typedef %s {"%(g.group(1)));
        return in_struct;

    print(ln,file=FO);
    return global_ln

def in_struct(ln):
    g = re.match(r"^}\s+PACK_SUFFIX\s+(\S+);",ln);
    if g:
        print("} __attribute__((packed)) %s;"%(g.group(1)));
        return global_ln;
    g = re.match(r"^}\s+(\S+);",ln);
    if g:
        print("} %s;"%(g.group(1)));
        return global_ln;

    ln = ln.replace("PACK_SUFFIX","__attribute__((packed))");
    ln = ln.replace("ib_gid_prefix_t","__be64");
    ln = ln.replace("ib_net64_t","__be64");
    ln = ln.replace("ib_net32_t","__be32");
    ln = ln.replace("ib_net16_t","__be16");
    ln = ln.replace("boolean_t","bool");
    print(ln)
    return in_struct;

mode = global_ln
with open(sys.argv[1]) as FI, open(sys.argv[2],"wt") as FO:
    for ln in FI:
        ln = ln.rstrip();
        mode = mode(ln);



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file
"""
Test module for pyverbs' mr module.
"""
from itertools import combinations as com
import random

from pyverbs.pyverbs_error import PyverbsRDMAError, PyverbsError
from pyverbs.tests.base import PyverbsTestCase
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.mr import MR, MW, DMMR
import pyverbs.tests.utils as u
import pyverbs.device as d
from pyverbs.pd import PD
import pyverbs.enums as e

MAX_IO_LEN = 1048576


class MRTest(PyverbsTestCase):
    """
    Test various functionalities of the MR class.
    """
    def test_reg_mr(self):
        """ Test ibv_reg_mr() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, u.get_mr_length(), f) as mr:
                        pass

    def test_dereg_mr(self):
        """ Test ibv_dereg_mr() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, u.get_mr_length(), f) as mr:
                        mr.close()

    @staticmethod
    def test_reg_mr_bad_flow():
        """ Verify that trying to register a MR with None PD fails """
        try:
            # Use the simplest access flags necessary
            MR(None, random.randint(0, 10000), e.IBV_ACCESS_LOCAL_WRITE)
        except TypeError as te:
            assert 'expected pyverbs.pd.PD' in te.args[0]
            assert 'got NoneType' in te.args[0]
        else:
            raise PyverbsRDMAErrno('Created a MR with None PD')

    def test_dereg_mr_twice(self):
        """ Verify that explicit call to MR's close() doesn't fails """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, u.get_mr_length(), f) as mr:
                        # Pyverbs supports multiple destruction of objects,
                        # we are not expecting an exception here.
                        mr.close()
                        mr.close()

    def test_reg_mr_bad_flags(self):
        """ Verify that illegal flags combination fails as expected """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                for i in range(5):
                    flags = random.sample([e.IBV_ACCESS_REMOTE_WRITE,
                                           e.IBV_ACCESS_REMOTE_ATOMIC],
                                          random.randint(1, 2))
                    mr_flags = 0
                    for i in flags:
                        mr_flags += i.value
                    try:
                        MR(pd, u.get_mr_length(), mr_flags)
                    except PyverbsRDMAError as err:
                        assert 'Failed to register a MR' in err.args[0]
                    else:
                        raise PyverbsRDMAError('Registered a MR with illegal falgs')

    def test_write(self):
        """
        Test writing to MR's buffer
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                for i in range(10):
                    mr_len = u.get_mr_length()
                    flags = u.get_access_flags(ctx)
                    for f in flags:
                        with MR(pd, mr_len, f) as mr:
                            write_len = min(random.randint(1, MAX_IO_LEN),
                                            mr_len)
                            mr.write('a' * write_len, write_len)

    def test_read(self):
        """
        Test reading from MR's buffer
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                for i in range(10):
                    mr_len = u.get_mr_length()
                    flags = u.get_access_flags(ctx)
                    for f in flags:
                        with MR(pd, mr_len, f) as mr:
                            write_len = min(random.randint(1, MAX_IO_LEN),
                                            mr_len)
                            write_str = 'a' * write_len
                            mr.write(write_str, write_len)
                            read_len = random.randint(1, write_len)
                            offset = random.randint(0, write_len-read_len)
                            read_str = mr.read(read_len, offset).decode()
                            assert read_str in write_str

    def test_lkey(self):
        """
        Test reading lkey property
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                length = u.get_mr_length()
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, length, f) as mr:
                        mr.lkey

    def test_rkey(self):
        """
        Test reading rkey property
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                length = u.get_mr_length()
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, length, f) as mr:
                        mr.rkey

    def test_buffer(self):
        """
        Test reading buf property
        """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                length = u.get_mr_length()
                flags = u.get_access_flags(ctx)
                for f in flags:
                    with MR(pd, length, f) as mr:
                        mr.buf


class MWTest(PyverbsTestCase):
    """
    Test various functionalities of the MW class.
    """
    def test_reg_mw_type1(self):
        """ Test ibv_alloc_mw() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with MW(pd, e.IBV_MW_TYPE_1):
                    pass

    def test_reg_mw_type2(self):
        """ Test ibv_alloc_mw() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with MW(pd, e.IBV_MW_TYPE_2):
                    pass

    def test_dereg_mw_type1(self):
        """ Test ibv_dealloc_mw() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with MW(pd, e.IBV_MW_TYPE_1) as mw:
                    mw.close()

    def test_dereg_mw_type2(self):
        """ Test ibv_dealloc_mw() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                with MW(pd, e.IBV_MW_TYPE_2) as mw:
                    mw.close()

    def test_reg_mw_wrong_type(self):
        """ Test ibv_alloc_mw() """
        for ctx, attr, attr_ex in self.devices:
            with PD(ctx) as pd:
                try:
                    mw_type = random.randint(3, 100)
                    MW(pd, mw_type)
                except PyverbsRDMAError:
                    pass
                else:
                    raise PyverbsError('Created a MW with type {t}'.\
                                       format(t=mw_type))


class DMMRTest(PyverbsTestCase):
    """
    Test various functionalities of the DMMR class.
    """
    def test_create_dm_mr(self):
        """
        Test ibv_reg_dm_mr
        """
        for ctx, attr, attr_ex in self.devices:
            if attr_ex.max_dm_size == 0:
                return
            with PD(ctx) as pd:
                for i in range(10):
                    dm_len = random.randrange(u.MIN_DM_SIZE, attr_ex.max_dm_size,
                                              u.DM_ALIGNMENT)
                    dm_attrs = u.get_dm_attrs(dm_len)
                    with d.DM(ctx, dm_attrs) as dm:
                        dm_mr_len = random.randint(1, dm_len)
                        dm_mr_offset = random.randint(0, (dm_len - dm_mr_len))
                        DMMR(pd, dm_mr_len, e.IBV_ACCESS_ZERO_BASED, dm=dm,
                             offset=dm_mr_offset)

    def test_destroy_dm_mr(self):
        """
        Test freeing of dm_mr
        """
        for ctx, attr, attr_ex in self.devices:
            if attr_ex.max_dm_size == 0:
                return
            with PD(ctx) as pd:
                for i in range(10):
                    dm_len = random.randrange(u.MIN_DM_SIZE, attr_ex.max_dm_size,
                                              u.DM_ALIGNMENT)
                    dm_attrs = u.get_dm_attrs(dm_len)
                    with d.DM(ctx, dm_attrs) as dm:
                        dm_mr_len = random.randint(1, dm_len)
                        dm_mr_offset = random.randint(0, (dm_len - dm_mr_len))
                        dm_mr = DMMR(pd, dm_mr_len, e.IBV_ACCESS_ZERO_BASED,
                                     dm=dm, offset=dm_mr_offset)
                        dm_mr.close()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file
"""
Test module for pyverbs' cq module.
"""
import random

from pyverbs.cq import CompChannel, CQ, CqInitAttrEx, CQEX
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.tests.base import PyverbsTestCase
import pyverbs.tests.utils as u
import pyverbs.enums as e


class CQTest(PyverbsTestCase):
    """
    Test various functionalities of the CQ class.
    """
    def test_create_cq(self):
        """
        Test ibv_create_cq()
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cqes = get_num_cqes(attr)
                comp_vector = int(ctx.num_comp_vectors * random.random())
                if random.choice([True, False]):
                    with CompChannel(ctx) as cc:
                        with CQ(ctx, cqes, None, cc, comp_vector):
                            pass
                else:
                    with CQ(ctx, cqes, None, None, comp_vector):
                        pass

    def test_create_cq_bad_flow(self):
        """
        Test ibv_create_cq() with a wrong comp_vector / cqe number
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cc = CompChannel(ctx)
                cqes = 100
                comp_vector = ctx.num_comp_vectors + int(100 *
                                                         random.random())
                has_cc = random.choice([True, False])
                if not has_cc:
                    cc = None
                try:
                    with CQ(ctx, cqes, None, cc, comp_vector):
                        pass
                except PyverbsError as e:
                    assert 'Failed to create a CQ' in e.args[0]
                    assert 'Invalid argument' in e.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQ with comp_vector={n} while device\'s num_comp_vectors={nc}'.
                        format(n=comp_vector, nc=ctx.num_comp_vectors))
                max_cqe = ctx.query_device().max_cqe
                cqes = random.randint(max_cqe + 1, max_cqe + 100)
                try:
                    with CQ(ctx, cqes, None, cc, 0):
                        pass
                except PyverbsError as err:
                    assert 'Failed to create a CQ' in err.args[0]
                    assert 'Invalid argument' in err.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQ with cqe={n} while device\'s max_cqe={nc}'.
                        format(n=cqes, nc=max_cqe))

    def test_destroy_cq(self):
        """
        Test ibv_destroy_cq()
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cqes = get_num_cqes(attr)
                comp_vector = int(ctx.num_comp_vectors * random.random())
                if random.choice([True, False]):
                    with CompChannel(ctx) as cc:
                        cq = CQ(ctx, cqes, None, cc, comp_vector)
                else:
                    cq = CQ(ctx, cqes, None, None, comp_vector)
                cq.close()


class CCTest(PyverbsTestCase):
    """
    Test various functionalities of the Completion Channel class.
    """
    def test_create_comp_channel(self):
        """
        Test ibv_create_comp_channel()
        """
        for ctx, attr, attr_ex in self.devices:
            with CompChannel(ctx):
                pass

    def test_destroy_comp_channel(self):
        """
        Test ibv_destroy_comp_channel()
        """
        for ctx, attr, attr_ex in self.devices:
            cc = CompChannel(ctx)
            cc.close()


class CQEXTest(PyverbsTestCase):
    """
    Test various functionalities of the CQEX class.
    """
    def test_create_cq_ex(self):
        """
        Test ibv_create_cq_ex()
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                with CQEX(ctx, get_attrs_ex(attr, attr_ex)):
                    pass

    def test_create_cq_ex_bad_flow(self):
        """
        Test ibv_create_cq_ex() with wrong comp_vector / number of cqes
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cq_attrs_ex = get_attrs_ex(attr, attr_ex)
                max_cqe = attr.max_cqe
                cq_attrs_ex.cqe = max_cqe + 1 + int(100 * random.random())
                try:
                    CQEX(ctx, cq_attrs_ex)
                except PyverbsError as e:
                    assert 'Failed to create extended CQ' in e.args[0]
                    assert ' Errno: 22' in e.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQEX with {c} CQEs while device\'s max CQE={dc}'.
                        format(c=cq_attrs_ex.cqe, dc=max_cqe))
                comp_channel = random.randint(ctx.num_comp_vectors, 100)
                cq_attrs_ex.comp_vector = comp_channel
                cq_attrs_ex.cqe = get_num_cqes(attr)
                try:
                    CQEX(ctx, cq_attrs_ex)
                except PyverbsError as e:
                    assert 'Failed to create extended CQ' in e.args[0]
                    assert ' Errno: 22' in e.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQEX with comp_vector={c} while device\'s num_comp_vectors={dc}'.
                        format(c=comp_channel, dc=ctx.num_comp_vectors))

    def test_destroy_cq_ex(self):
        """
        Test ibv_destroy_cq() for extended CQs
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                with CQEX(ctx, get_attrs_ex(attr, attr_ex)) as cq:
                    cq.close()

def get_num_cqes(attr):
    max_cqe = attr.max_cqe
    return int((max_cqe + 1) * random.random())


def get_attrs_ex(attr, attr_ex):
    cqe = get_num_cqes(attr)
    wc_flags = list(e.ibv_create_cq_wc_flags)
    # Flow tag is not always supported, doesn't have a capability bit to check
    wc_flags.remove(e.IBV_WC_EX_WITH_FLOW_TAG)
    if attr_ex.tm_caps.max_ops == 0:
        wc_flags.remove(e.IBV_WC_EX_WITH_TM_INFO)
    if attr_ex.raw_packet_caps & e.IBV_RAW_PACKET_CAP_CVLAN_STRIPPING == 0:
        wc_flags.remove(e.IBV_WC_EX_WITH_CVLAN)
    sample = u.sample(wc_flags)
    wc_flags = 0
    for flag in sample:
        wc_flags |= flag
    comp_mask = random.choice([0, e.IBV_CQ_INIT_ATTR_MASK_FLAGS])
    flags = 0
    if comp_mask is not 0:
        attr_flags = list(e.ibv_create_cq_attr_flags)
        sample = u.sample(attr_flags)
        for flag in sample:
            flags |= flag
    return CqInitAttrEx(cqe=cqe, wc_flags=wc_flags, comp_mask=comp_mask,
                        flags=flags)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2018, Mellanox Technologies. All rights reserved.  See COPYING file
import unittest,os,os.path,fnmatch
import tests


def test_all():
    # FIXME: This implementation is for older Python versions, will
    # be replaced with discover()
    return test_suite

module = __import__("tests")
fns = [os.path.splitext(I)[0] for I in fnmatch.filter(os.listdir(module.__path__[0]),"*.py")]
fns.remove("__init__")
for I in fns:
    __import__("tests." + I)
test_suite = unittest.TestSuite(unittest.defaultTestLoader.loadTestsFromNames(fns,module))

if __name__ == '__main__':
    unittest.main(defaultTest="test_all")





# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019, Mellanox Technologies. All rights reserved.

#cython: language_level=3

from pyverbs.device cimport Context
cimport pyverbs.libibverbs as v
from .base cimport PyverbsCM


cdef class PD(PyverbsCM):
    cdef v.ibv_pd *pd
    cdef Context ctx
    cdef add_ref(self, obj)
    cdef object mrs
    cdef object mws
    cdef object ahs
    cdef object qps
